//
// Copyright (c) 2014 CNRS
// Authors: Florent Lamiraux
//
// This file is part of hpp-core
// hpp-core is free software: you can redistribute it
// and/or modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation, either version
// 3 of the License, or (at your option) any later version.
//
// hpp-core is distributed in the hope that it will be
// useful, but WITHOUT ANY WARRANTY; without even the implied warranty
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Lesser Public License for more details.  You should have
// received a copy of the GNU Lesser General Public License along with
// hpp-core  If not, see
// <http://www.gnu.org/licenses/>.

#include <hpp/util/debug.hh>
#include <hpp/model/device.hh>
#include <hpp/model/joint.hh>
#include <hpp/model/joint-configuration.hh>
#include <hpp/core/config-projector.hh>
#include <hpp/core/straight-path.hh>

namespace hpp {
  namespace core {
    StraightPath::StraightPath (const DevicePtr_t& device,
				ConfigurationIn_t init,
				ConfigurationIn_t end,
				value_type length) :
      parent_t (interval_t (0, length), device->configSize (),
		device->numberDof ()),
      device_ (device), initial_ (init), end_ (end)
    {
      assert (length >= 0);
    }

    StraightPath::StraightPath (const StraightPath& path) :
      parent_t (path), initial_ (path.initial_), end_ (path.end_)
    {
    }

    bool StraightPath::impl_compute (ConfigurationOut_t result,
				     value_type param) const
    {
      if (param == timeRange ().first || timeRange ().second == 0) {
	result = initial_;
	return true;
      }
      if (param == timeRange ().second) {
	result = end_;
	return true;
      }
      value_type u = param/timeRange ().second;
      if (timeRange ().second == 0)
	u = 0;
      // Loop over device joint and interpolate
      const JointVector_t& jv (device_->getJointVector ());
      for (model::JointVector_t::const_iterator itJoint = jv.begin ();
	   itJoint != jv.end (); ++itJoint) {
	std::size_t rank = (*itJoint)->rankInConfiguration ();
	(*itJoint)->configuration ()->interpolate
	  (initial_, end_, u, rank, result);
      }

      const ConstraintSetPtr_t& cs (constraints ());
      if (cs) {
	if (cs->configProjector ()) {
	  cs->configProjector ()->rightHandSideFromConfig (initial_);
	}
        return cs->apply (result);
      }
      return true;
    }
    PathPtr_t StraightPath::extract (const interval_t& subInterval) const
    {
      // Length is assumed to be proportional to interval range
      value_type l = fabs (subInterval.second - subInterval.first);

      Configuration_t q1 ((*this) (subInterval.first));
      Configuration_t q2 ((*this) (subInterval.second));
      PathPtr_t result = StraightPath::create (device_, q1, q2, l);
      result->constraints (this->constraints ());
      return result;
    }

    DevicePtr_t StraightPath::device () const
    {
      return device_;
    }
  } //   namespace core
} // namespace hpp

